import sys
from pathlib import Path
from csv import DictReader
import time
import multiprocessing as mp

import sqlalchemy as sa
import pandas as pd
import geopandas as gpd
import numpy as np
import yaml

from .util import count_lines, prepare_db, printProgressBar

dummy_date = "19000101"


def load_parcel_info(dsn, counts_file, shape_file, table_name):
    """Loads the parcel info from the

    :param dsn: Data source name where to write to
    counts_file: CSV file from which pixel counts should be read
    shape_file: shapefile whose .DBF file should be used as parcel info.
    table_name: name of the table to write parcel info into

    :return:
    """
    shp_fname = Path(shape_file)
    df = gpd.read_file(shp_fname)
    df["area_ha"] = df.geometry.area/1e4
    df = df.set_index("fieldid")

    fname_counts = Path(counts_file)
    df_counts = pd.read_csv(fname_counts)
    df_counts.set_index("field_ID", inplace=True)
    df["pixcount"] = df_counts["pixcount"]

    df_out = pd.DataFrame({"fieldID": df.index,
                           "year": df.year.astype(np.int32),
                           "pixcount": df.pixcount.astype(np.int32),
                           "area_ha": df.area_ha,
                           "cat_gewasc": df.cat_gewasc.apply(str),
                           "gws_gewasc": df.gws_gewasc.astype(np.int32),
                           "gws_gewas": df.gws_gewas.apply(str),
                           "provincie": df.provincie.apply(str),
                           "gemeente": df.gemeente.apply(str),
                           "regio": df.regio.apply(str),
                           "pc4": df.PC4.apply(str),
                           "woonplaats": df.woonplaats.apply(str),
                           "waterschap": df.waterschap.apply(str),
                           })

    engine = sa.create_engine(dsn)
    df_out.to_sql(table_name, engine, if_exists="replace", index=False)


def process_rows(rows):
    df = pd.DataFrame()
    fieldIDs = []
    for column_name, row in rows.items():
        fieldIDs.append(int(row.pop("field_ID")))
        count = row.pop("count")
        recs = []
        for sdate, value in row.items():
            value = float(value)
            if value == 0.:
                continue
            recs.append({"day": sdate, "value": float(value), "band": column_name})
        if not recs:  # only zero (null) values for the column
            # We add one dummy record to make sure we can create the dataframe properly
            recs.append({"day": dummy_date, "value": None, "band": column_name})
        df_tmp = pd.DataFrame(recs)
        try:
            df_tmp["day"] = pd.to_datetime(df_tmp.day).dt.date
        except:
            pass
        df = pd.concat([df, df_tmp])
    df = df.pivot(index="day", columns="band", values="value")
    df.reset_index(inplace=True)

    if len(set(fieldIDs)) > 1:
        msg = f"FieldIDs are not the same for this row: {fieldIDs}"
        raise RuntimeError(msg)
    df["fieldID"] = fieldIDs[0]

    ix = (df.day == pd.to_datetime(dummy_date))
    if any(ix):
        df = df[~ix]

    return df


def write_to_database(engine, table_name, mean_csv_readers, nlines):
    # printProgressBar(0, nlines-1, prefix='Progress:', suffix='Complete', length=50)
    this_line = 0
    while True:
        try:
            rows = {column_name: next(reader) for column_name, reader in mean_csv_readers.items()}
            df = process_rows(rows)
            df.to_sql(table_name, engine, if_exists="append", index=False)
            this_line += 1
        #     if this_line % 1000 == 0:
        #         printProgressBar(this_line, nlines, prefix='Progress:', suffix='Complete', length=50)
        except StopIteration:
            break
        except Exception as e:
            pass


def load_satellite_csv(dataset_name, dsn, bands):
    nlines = count_lines(bands)  # 803016
    mean_csv_readers = {}
    for column_name, csv_fname in bands.items():
        mean_csv_readers[column_name] = DictReader(open(csv_fname))
    t1 = time.time()
    engine = prepare_db(dsn, table_name=dataset_name, bands=mean_csv_readers.keys())
    write_to_database(engine, dataset_name, mean_csv_readers, nlines)


def load_data(yaml_file):
    """Loads data point to by the YAML config file.

    :param yaml_file:
    :return:
    """

    grompy_conf = yaml.safe_load(open(yaml_file))
    parcel_info = grompy_conf.pop("parcel_info")
    load_parcel_info(**parcel_info)

    process_list = []
    for dataset_name, description in grompy_conf["datasets"].items():
        print(f"Starting loading of: {dataset_name}")
        p = mp.Process(target=load_satellite_csv, args=(dataset_name,), kwargs=description)
        process_list.append(p)
    for p in process_list:
        p.start()

    while any([p.is_alive() for p in process_list]):
        try:
            print("Data loading running.....")
            time.sleep(10)
        except KeyboardInterrupt:
            for p in process_list:
                p.kill()
            sys.exit()

